//******************************************************************************************
//* Wada Basin, by Paul Nylander, 6/19/05
//******************************************************************************************

// In order to compile this, you might need these OpenGL files: http://www.ics.uci.edu/~frost/past/ics186a/EssentialFiles.zip
// Simply place the files in these folders: http://www.ics.uci.edu/~frost/past/ics186a/wheretheygo.html

#include <math.h> // sqrt, pow
#include <windows.h> // min, max
#include <GL/glut.h>

//Basic Functions
#define pi 3.14159265358979323846
double infinity=1.0e308;
double rem(double x, double y) {int i=x/y; return (x-i*y);}
double sqr(double x) {return x*x;}
double seed=pi;
double random() {seed=rem(1000*pi*seed,1); return seed;} // quick random

int xscreen=280,yscreen=280,window;

//Image Structure
struct Image{int xsize, ysize; char *pixels;};
typedef struct Image Image;
Image *screen;

//******************************************************************************************
//* 3D Point Class
//******************************************************************************************

class pt {public: double x,y,z;
 pt():x(0.0),y(0.0),z(0.0) {}
 pt(const double x1, const double y1, const double z1=0.0) {x=x1; y=y1; z=z1;}
 pt(const pt& p):x(p.x),y(p.y),z(p.z) {}
 pt& operator=(const pt& p) {x=p.x; y=p.y; z=p.z; return *this;}
 pt& operator+=(const pt& p) {x+=p.x; y+=p.y; z+=p.z; return *this;}
 pt& operator-=(const pt& p) {x-=p.x; y-=p.y; z-=p.z; return *this;}

 friend pt operator+(const pt& p1, const pt& p2) {return pt(p1.x+p2.x,p1.y+p2.y,p1.z+p2.z);}
 friend pt operator+(const double& x, const pt& p) {return pt(x+p.x,x+p.y,x+p.z);}
 friend pt operator+(const pt& p, const double& x) {return x+p;}
 friend pt operator-(const pt& p) {return pt(-p.x,-p.y,-p.z);}
 friend pt operator-(const pt& p1, const pt& p2) {return p1+(-p2);}
 friend pt operator-(const double& x, const pt& p) {return x+(-p);}
 friend pt operator-(const pt& p, const double& x) {return p+(-x);}
 friend double operator*(const pt& p1, const pt& p2) { // dot product
  return p1.x*p2.x+p1.y*p2.y+p1.z*p2.z;
 }
 friend pt operator*(const double& x, const pt& p) {return pt(x*p.x,x*p.y,x*p.z);}
 friend pt operator*(const pt& p, const double& x) {return x*p;}
 friend pt operator/(const pt& p, const double& x) {return pt(p.x/x,p.y/x,p.z/x);}
 friend pt operator/(const pt& p, const int n) {return p/double(n);}
 friend bool operator==(const pt& p1, const pt& p2) {pt p=p2-p1; return sqrt(p*p)<0.001;}
 friend bool operator!=(const pt& p1, const pt& p2) {return !(p1==p2);}
 friend double norm(const pt& p) {return p*p;}
 friend double length(const pt& p) {return sqrt(p*p);}
 friend pt normalize(const pt& p) {double d=length(p); return p/d;}
 friend double distance(const pt& p1, const pt& p2) {return length(p2-p1);}
 friend pt cross(const pt& p1, const pt& p2) {return pt(p1.y*p2.z-p1.z*p2.y, p1.z*p2.x-p1.x*p2.z, p1.x*p2.y-p1.y*p2.x);}
};

//******************************************************************************************
//* Red, Green, Blue (RGB) Color Class
//******************************************************************************************

class rgb{public: double r,g,b;
 rgb():r(0.0),g(0.0),b(0.0) {}
 rgb(const double r1, const double g1, const double b1) {r=r1; g=g1; b=b1;}
 rgb(const double x) {r=g=b=x;}
 rgb(const rgb& c):r(c.r),g(c.g),b(c.b) {}
 rgb& operator=(const rgb& c) {r=c.r; g=c.g; b=c.b; return *this;}
 rgb& operator+=(const rgb& c) {r+=c.r; g+=c.g; b+=c.b; return *this;}
 rgb& operator-=(const rgb& c) {r-=c.r; g-=c.g; b-=c.b; return *this;}
 friend rgb operator+(const rgb& c1, const rgb& c2) {return rgb(c1.r+c2.r,c1.g+c2.g,c1.b+c2.b);}
 friend rgb operator-(const rgb& c) {return rgb(-c.r,-c.g,-c.b);}
 friend rgb operator-(const rgb& c1, const rgb& c2) {return c1+(-c2);}
 friend rgb operator*(const double& x, const rgb& c) {return rgb(x*c.r,x*c.g,x*c.b);}
 friend rgb operator*(const rgb& c, const double& x) {return x*c;}
 friend rgb operator*(const rgb& c, const int& x) {return double(x)*c;}
 friend rgb operator*(const int& x, const rgb& c) {return double(x)*c;}
 friend rgb operator*(const rgb& c1, const rgb& c2) {return rgb(c1.r*c2.r,c1.g*c2.g,c1.b*c2.b);} // just for fun
 friend rgb operator/(const rgb& c, const double& x) {return rgb(c.r/x,c.g/x,c.b/x);}
 friend rgb operator/(const rgb& c, const int n) {return c/double(n);}
 friend bool operator==(const rgb& c1, const rgb& c2) {return (c1.r==c2.r && c1.g==c2.g && c1.b==c2.b);}
 friend bool operator!=(const rgb& c1, const rgb& c2) {return (c1.r!=c2.r && c1.g!=c2.g && c1.b!=c2.b);}
};

char cast(double x) {int i=int(min(255,256*min(1,max(0,x)))); return i;}
void pixel(int x, int y, rgb c, Image *image) {if(x>=0 && x<image->xsize && y>=0 && y<image->ysize) {int k=3*(y*image->xsize+x);
 image->pixels[k]=cast(c.r); image->pixels[k+1]=cast(c.g); image->pixels[k+2]=cast(c.b);
}}

//******************************************************************************************
//* Ray Tracer
//******************************************************************************************

//Global Constants
int MaxTraceLevel=50;

//Camera
pt camloc=pt(0,0,0.5);
pt lookat=pt(0,0,0);
pt camdir=normalize(lookat-camloc);
pt camright=normalize(cross(camdir,pt(0,1,0)));
pt camup=normalize(cross(camright,camdir));

//Lights
class light {public: pt p; rgb color;
 light():p(0,0,0),color(1,1,1) {}
 light(pt p1, rgb c1) {p=p1; color=c1;}
 light(const light& light1):p(light1.p),color(light1.color) {}
 light& operator=(const light& light1) {p=light1.p; color=light1.color; return *this;}
};
int nlight=1;
light lights[1]={light(pt(0,0,1), 0*rgb(1,1,1))};

//Objects
double precision=1.0e-10;
enum types{sphere}; // only spheres are supported so far
class object {public: int type; pt p; double size, ambient,reflectivity,shininess; rgb color;
 object() {type=0; p=pt(0,0,0); size=1; color=rgb(1,1,1); ambient=0; reflectivity=0; shininess=0;}
 object(int type1, pt p1, double size1, rgb c1, double ambient1, double reflectivity1, double shininess1) {
  type=type1; p=p1; size=size1; color=c1; ambient=ambient1; reflectivity=reflectivity1; shininess=shininess1;
 }
 object(const object& object1) {
  type=object1.type; p=object1.p; size=object1.size; color=object1.color;
  ambient=object1.ambient; reflectivity=object1.reflectivity; shininess=object1.shininess;
 }
 object& operator=(const object& object1) {
  type=object1.type; p=object1.p; size=object1.size; color=object1.color;
  ambient=object1.ambient; reflectivity=object1.reflectivity; shininess=object1.shininess;
  return *this;
 }
 friend double MinIntersection(object& object1, pt& p, pt& dir) {pt v=p-object1.p;
  double a=1, b=2*(v*dir), c=norm(v)-sqr(object1.size), d=sqr(b)-4*a*c;
  if(d<0) return 0; else {d=sqrt(d); double d1=(-b-d)/(2*a), d2=(-b+d)/(2*a); // shouldn't be negative!
   if(d1<precision && d2<precision) return 0;
   if(d1<precision) return d2;
   if(d2<precision) return d1;
   else return min(d1,d2); // min2
  }
 }
 friend pt GetNormal(object& object1, pt& p) {return normalize(p-object1.p);}
};
int nobject=4;
double r=1, ambient=0.25, reflectivity=1, shininess=10;
object objects[4]={
 object(sphere,pt(1,sqrt(1.0/3),0),r,rgb(1,0,0),ambient,reflectivity,shininess),
 object(sphere,pt(-1,sqrt(1.0/3),0),r,rgb(0,1,0),ambient,reflectivity,shininess),
 object(sphere,pt(0,-sqrt(4.0/3),0),r,rgb(0,0,1),ambient,reflectivity,shininess),
 object(sphere,pt(0,0,-sqrt(8.0/3)),r,rgb(1,1,1),ambient,reflectivity,shininess)
};

//Ray Tracer
rgb RayTrace(pt, pt, int);
rgb GetColor(object object1, pt p, pt dir, int TraceLevel) {double t1;
 rgb color=object1.ambient*object1.color;
 if(TraceLevel<MaxTraceLevel) {
  pt normal=GetNormal(object1,p), reflect=normalize(dir-2*(normal*dir)*normal);
  pt lightdir,v,shadowdir; double intensity,d,specular; bool shadowed;
  if(object1.reflectivity>0) color+=object1.reflectivity*RayTrace(p,reflect,TraceLevel+1);
  for(int ilight=0; ilight<nlight; ilight++) {
   lightdir=normalize(lights[ilight].p-p); intensity=normal*lightdir;
   if(intensity>0) {
    v=lights[ilight].p-p; shadowdir=normalize(v); d=length(v); shadowed=0;
    for(int iobject=0; iobject<nobject && !shadowed; iobject++) {
     t1=MinIntersection(objects[iobject],p,shadowdir);
     if(t1!=0 && t1<d) shadowed=1;
    }
    if(!shadowed) {
     color+=intensity*lights[ilight].color;
     specular=reflect*lightdir;
     color+=0.6*pow(specular,object1.shininess)*lights[ilight].color;
    }
   }
  }
 }
 return color;
}
rgb RayTrace(pt p, pt dir, int TraceLevel) {double dmin=infinity,d; object object1;
 for(int iobject=0; iobject<nobject; iobject++) {
  d=MinIntersection(objects[iobject], p, dir);
  if(d!=0 && d<dmin) {dmin=d; object1=objects[iobject];};
 }
 return (dmin<infinity)?GetColor(object1, p+dmin*dir, dir, TraceLevel):rgb(0);
}

//******************************************************************************************
//* OpenGL Display
//******************************************************************************************

void display(void) {
 glLoadIdentity();
 glClearColor(0,0,0,1); glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 gluOrtho2D(0,xscreen,0,yscreen);
 glRasterPos2i(0,0);
 glDrawPixels(screen->xsize, screen->ysize, GL_RGB, GL_UNSIGNED_BYTE, screen->pixels);
 glutSwapBuffers();
}

void keyboard(unsigned char key, int xmouse, int ymouse) {switch(key) {
 case 'q': case 'Q': case 27: glutDestroyWindow(window); exit(0); break;
}}

void main(int argc, char **argv) {
 glutInit(&argc, argv);
 glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
 glutInitWindowSize(xscreen, yscreen);
 window=glutCreateWindow("Wada Basin");
 glutDisplayFunc(display);
 glutKeyboardFunc(keyboard);
 screen=new Image();
 screen->xsize=xscreen; screen->ysize=yscreen;
 screen->pixels=new char[3*(xscreen+1)*yscreen];
 for(int i=0; i<yscreen; i++) {double y=(2.0*i)/yscreen-1;
  for(int j=0; j<xscreen; j++) {double x=(2.0*j)/xscreen-1;
   pixel(j,i,RayTrace(camloc,normalize(camdir+x*camright+y*camup),1),screen);
  }
 }
 display();
 glutMainLoop();
}
